import pytest
from sqlalchemy import String
from sqlalchemy.orm import Mapped, Session, mapped_column


@pytest.mark.database
@pytest.mark.integration
def test_main(session: Session, init_db: None) -> None:
    from mixemy import models, repositories, schemas, services

    class ItemModel(models.IdAuditModel):
        __table_args__ = {"extend_existing": True}  # noqa: RUF012
        value: Mapped[str] = mapped_column(String)
        nullable_value: Mapped[str | None] = mapped_column(String, nullable=True)

    class ItemInput(schemas.InputSchema):
        value: str

    class ItemUpdate(ItemInput):
        nullable_value: str | None

    class ItemFilter(schemas.InputSchema):
        value: list[str]

    class ItemOutput(schemas.IdAuditOutputSchema):
        value: str
        nullable_value: str | None

    class ItemRepository(repositories.IdAuditSyncRepository[ItemModel]):
        model_type = ItemModel

    class ItemService(
        services.IdAuditSyncService[
            ItemModel, ItemInput, ItemUpdate, ItemFilter, ItemOutput
        ]
    ):
        repository_type = ItemRepository
        output_schema_type = ItemOutput

    item_service = ItemService()

    test_one = ItemInput(value="test_one")
    test_two = ItemInput(value="test_two")
    test_three = ItemInput(value="test_one")
    test_one_update = ItemUpdate(value="test_one", nullable_value="test_one_updated")
    test_one_id = None

    item_one = item_service.create(db_session=session, object_in=test_one)
    item_two = item_service.create(db_session=session, object_in=test_two)
    item_service.create(db_session=session, object_in=test_three)

    test_one_id = item_one.id

    assert item_one.value == "test_one"
    assert item_two.value == "test_two"

    item_one = item_service.read(db_session=session, id=item_one.id)
    item_two = item_service.read(db_session=session, id=item_two.id)

    assert item_one is not None
    assert item_two is not None
    assert item_one.value == "test_one"
    assert item_one.nullable_value is None
    assert item_two.value == "test_two"
    assert item_two.nullable_value is None

    item_one = item_service.update(
        db_session=session, id=item_one.id, object_in=test_one_update
    )

    assert item_one.value == "test_one"
    assert item_one.nullable_value == "test_one_updated"
    assert item_one.id == test_one_id

    items = item_service.read_multi(
        db_session=session, filters=ItemFilter(value=["test_one"])
    )

    assert len(items) == 2

    item_service.delete(db_session=session, id=item_one.id)

    item_one = item_service.read(db_session=session, id=item_one.id)
    item_two = item_service.read(db_session=session, id=item_two.id)

    assert item_one is None
    assert item_two is not None
    assert item_two.value == "test_two"

    item_service.delete(db_session=session, id=item_two.id)

    item_two = item_service.read(db_session=session, id=item_two.id)

    assert item_two is None



from ._audit import AuditSyncRepository
from ._base import BaseSyncRepository
from ._id import IdSyncRepository
from ._id_audit import IdAuditSyncRepository

__all__ = [
    "AuditSyncRepository",
    "BaseSyncRepository",
    "IdAuditSyncRepository",
    "IdSyncRepository",
]



from abc import ABC

from mixemy.services._base._base import BaseService
from mixemy.types import (
    AuditModelType,
    CreateSchemaType,
    FilterSchemaType,
    OutputSchemaType,
    UpdateSchemaType,
)


class IdService(
    BaseService[
        AuditModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    pass



from ._id import ID, PASSWORD_HASH, PASSWORD_INPUT
from ._models import AuditModelType, BaseModelType, IdAuditModelType, IdModelType
from ._repositories import ReturnOrRaiseType
from ._schemas import (
    AuditPaginationSchemaType,
    CreateSchemaType,
    FilterSchemaType,
    OutputSchemaType,
    PaginationSchemaType,
    UpdateSchemaType,
)
from ._session import SessionType

__all__ = [
    "ID",
    "PASSWORD_HASH",
    "PASSWORD_INPUT",
    "AuditModelType",
    "AuditPaginationSchemaType",
    "BaseModelType",
    "CreateSchemaType",
    "FilterSchemaType",
    "IdAuditModelType",
    "IdModelType",
    "OutputSchemaType",
    "PaginationSchemaType",
    "ReturnOrRaiseType",
    "SessionType",
    "UpdateSchemaType",
]



from abc import ABC

from sqlalchemy.orm import Session

from mixemy.repositories.sync import BaseSyncRepository
from mixemy.services._base import BaseService
from mixemy.types import (
    BaseModelType,
    CreateSchemaType,
    FilterSchemaType,
    OutputSchemaType,
    PaginationSchemaType,
    UpdateSchemaType,
)


class BaseSyncService(
    BaseService[
        BaseModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    repository_type: type[BaseSyncRepository[BaseModelType]]

    def __init__(self) -> None:
        super().__init__()
        self.repository = self.repository_type()
        self.model = self.repository.model

    def create(self, db_session: Session, object_in: CreateSchemaType) -> BaseModelType:
        return self.repository.create(
            db_session=db_session, db_object=self._to_model(schema=object_in)
        )

    def read_multi(
        self,
        db_session: Session,
        filters: FilterSchemaType | None = None,
        pagination: PaginationSchemaType | None = None,
    ) -> list[OutputSchemaType]:
        return [
            self._to_schema(model=model)
            for model in self.repository.read_multi(
                db_session=db_session, before_filter=filters, after_filter=pagination
            )
        ]



from abc import ABC
from typing import Generic

from mixemy._exceptions import MixemyServiceSetupError
from mixemy.types import (
    BaseModelType,
    CreateSchemaType,
    FilterSchemaType,
    OutputSchemaType,
    UpdateSchemaType,
)
from mixemy.utils import to_model, to_schema


class BaseService(
    Generic[
        BaseModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    output_schema_type: type[OutputSchemaType]
    model: type[BaseModelType]

    def __init__(self) -> None:
        self._verify_init()
        self.output_schema = self.output_schema_type

    def _to_model(self, schema: CreateSchemaType | UpdateSchemaType) -> BaseModelType:
        return to_model(schema=schema, model=self.model)

    def _to_schema(self, model: BaseModelType) -> OutputSchemaType:
        return to_schema(model=model, schema=self.output_schema)

    def _verify_init(self) -> None:
        for field in ["output_schema_type", "repository_type"]:
            if not hasattr(self, field):
                raise MixemyServiceSetupError(service=self, undefined_field=field)



from abc import ABC

from mixemy.repositories.asyncio import IdAuditAsyncRepository
from mixemy.services.asyncio._audit import AuditAsyncService
from mixemy.services.asyncio._id import IdAsyncService
from mixemy.types import (
    CreateSchemaType,
    FilterSchemaType,
    IdAuditModelType,
    OutputSchemaType,
    UpdateSchemaType,
)


class IdAuditAsyncService(
    IdAsyncService[
        IdAuditModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    AuditAsyncService[
        IdAuditModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    repository_type: type[IdAuditAsyncRepository[IdAuditModelType]]  # pyright: ignore[reportIncompatibleVariableOverride] - https://github.com/python/typing/issues/548
    repository: IdAuditAsyncRepository[IdAuditModelType]  # pyright: ignore[reportIncompatibleVariableOverride] - https://github.com/python/typing/issues/548



from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import DeclarativeBase


class BaseModel(DeclarativeBase):
    __abstract__ = True

    @declared_attr  # pyright: ignore[reportArgumentType]
    @classmethod
    def __tablename__(cls) -> str:
        return cls.__name__.lower()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.__dict__}>"



from ._audit import AuditAsyncRepository
from ._base import BaseAsyncRepository
from ._id import IdAsyncRepository
from ._id_audit import IdAuditAsyncRepository

__all__ = [
    "AuditAsyncRepository",
    "BaseAsyncRepository",
    "IdAsyncRepository",
    "IdAuditAsyncRepository",
]



from abc import ABC
from typing import Any, Generic

from sqlalchemy import Select

from mixemy._exceptions import MixemyRepositorySetupError
from mixemy.schemas import InputSchema
from mixemy.types import (
    BaseModelType,
    PaginationSchemaType,
)
from mixemy.utils import unpack_schema


class BaseRepository(Generic[BaseModelType], ABC):
    model_type: type[BaseModelType]

    def __init__(self) -> None:
        self._verify_init()
        self.model = self.model_type

    @staticmethod
    def _add_after_filter(
        statement: Select[Any], filter: PaginationSchemaType | None
    ) -> Select[Any]:
        if filter is not None:
            statement = statement.offset(filter.offset).limit(filter.limit)

        return statement

    def _add_before_filter(
        self, statement: Select[Any], filter: InputSchema | None
    ) -> Select[Any]:
        if filter is not None:
            for item, value in unpack_schema(schema=filter).items():
                if isinstance(value, list):
                    statement = statement.where(getattr(self.model, item).in_(value))
                else:
                    statement = statement.where(getattr(self.model, item) == value)

        return statement

    @staticmethod
    def _update_db_object(db_object: BaseModelType, object_in: InputSchema) -> None:
        for field, value in unpack_schema(schema=object_in).items():
            if hasattr(db_object, field):
                setattr(db_object, field, value)

    def _verify_init(self) -> None:
        for field in ["model_type"]:
            if not hasattr(self, field):
                raise MixemyRepositorySetupError(repository=self, undefined_field=field)



from abc import ABC
from typing import Any, override

from sqlalchemy import Select, asc, desc

from mixemy.repositories._base._base import BaseRepository
from mixemy.schemas.paginations import AuditPaginationFilter, OrderDirection
from mixemy.types import (
    AuditModelType,
    PaginationSchemaType,
)


class AuditRepository(BaseRepository[AuditModelType], ABC):
    @override
    @staticmethod
    def _add_after_filter(
        statement: Select[Any], filter: PaginationSchemaType | None
    ) -> Select[Any]:
        if filter is not None:
            statement = statement.offset(filter.offset).limit(filter.limit)
        if isinstance(filter, AuditPaginationFilter):
            match filter.order_direction:
                case OrderDirection.ASC:
                    statement = statement.order_by(asc(filter.order_by))
                case OrderDirection.DESC:
                    statement = statement.order_by(desc(filter.order_by))

        return statement



from typing import TYPE_CHECKING, TypeVar

# from mixemy.models import AuditModel, BaseModel, IdAuditModel, IdModel

if TYPE_CHECKING:
    from mixemy.models import AuditModel, BaseModel, IdAuditModel, IdModel

BaseModelType = TypeVar("BaseModelType", bound="BaseModel")
IdModelType = TypeVar("IdModelType", bound="IdModel")
AuditModelType = TypeVar("AuditModelType", bound="AuditModel")
IdAuditModelType = TypeVar("IdAuditModelType", bound="IdAuditModel")



from mixemy.schemas.paginations._base import PaginationFilter
from mixemy.schemas.paginations._order_enums import OrderBy, OrderDirection


class AuditPaginationFilter(PaginationFilter):
    order_by: OrderBy = OrderBy.CREATED_AT
    order_direction: OrderDirection = OrderDirection.DESC



from abc import ABC
from collections.abc import Sequence
from typing import Any

from sqlalchemy import Select, func, select
from sqlalchemy.orm import Session

from mixemy.repositories._base import BaseRepository
from mixemy.schemas import InputSchema
from mixemy.types import (
    BaseModelType,
    PaginationSchemaType,
)


class BaseSyncRepository(BaseRepository[BaseModelType], ABC):
    def create(self, db_session: Session, db_object: BaseModelType) -> BaseModelType:
        return self._add_refreshing(db_session=db_session, db_object=db_object)

    def read_multi(
        self,
        db_session: Session,
        after_filter: PaginationSchemaType | None = None,
        before_filter: InputSchema | None = None,
    ) -> Sequence[BaseModelType]:
        return self._execute_returning_all(
            db_session=db_session,
            statement=self._add_after_filter(
                statement=self._add_before_filter(
                    statement=select(self.model), filter=before_filter
                ),
                filter=after_filter,
            ),
        )

    def update(
        self,
        db_session: Session,
        db_object: BaseModelType,
        object_in: InputSchema,
    ) -> BaseModelType:
        self._update_db_object(db_object=db_object, object_in=object_in)
        return self._add_refreshing(db_session=db_session, db_object=db_object)

    def delete(self, db_session: Session, db_object: BaseModelType) -> None:
        self._delete_refreshing(db_session=db_session, db_object=db_object)

    def count(
        self,
        db_session: Session,
        before_filter: InputSchema | None = None,
    ) -> int:
        return self._execute_returning_one(
            db_session=db_session,
            statement=self._add_before_filter(
                statement=select(func.count()).select_from(self.model),
                filter=before_filter,
            ),
        )

    @staticmethod
    def _execute_returning_all(
        db_session: Session, statement: Select[Any]
    ) -> Sequence[BaseModelType]:
        return db_session.execute(statement=statement).scalars().all()

    @staticmethod
    def _execute_returning_one(db_session: Session, statement: Select[Any]) -> Any:
        return db_session.execute(statement=statement).scalar_one()

    @staticmethod
    def _add_refreshing(db_session: Session, db_object: BaseModelType) -> BaseModelType:
        db_session.add(db_object)
        db_session.commit()
        db_session.refresh(db_object)

        return db_object

    @staticmethod
    def _delete_refreshing(db_session: Session, db_object: BaseModelType) -> None:
        db_session.delete(db_object)
        db_session.commit()
        db_session.refresh(db_object)



from abc import ABC

from mixemy.repositories._base import AuditRepository
from mixemy.repositories.asyncio._base import BaseAsyncRepository
from mixemy.types import AuditModelType


class AuditAsyncRepository(
    BaseAsyncRepository[AuditModelType], AuditRepository[AuditModelType], ABC
):
    pass



from abc import ABC

from mixemy.repositories.asyncio._audit import AuditAsyncRepository
from mixemy.repositories.asyncio._id import IdAsyncRepository
from mixemy.types import IdAuditModelType


class IdAuditAsyncRepository(
    IdAsyncRepository[IdAuditModelType],
    AuditAsyncRepository[IdAuditModelType],
    ABC,
):
    pass



from abc import ABC

from mixemy.repositories.sync._audit import AuditSyncRepository
from mixemy.repositories.sync._id import IdSyncRepository
from mixemy.types import IdAuditModelType


class IdAuditSyncRepository(
    IdSyncRepository[IdAuditModelType],
    AuditSyncRepository[IdAuditModelType],
    ABC,
):
    pass



from datetime import UTC, datetime
from typing import overload


@overload
def serialize_datetime(data: datetime) -> str: ...


@overload
def serialize_datetime(data: None) -> None: ...


def serialize_datetime(data: datetime | None) -> str | None:
    if data is None:
        return data

    if not data.tzinfo:
        data = data.replace(tzinfo=UTC)

    return data.isoformat()



from collections.abc import Callable
from typing import overload

from pydantic import SecretStr


@overload
def serialize_secretstr(
    data: SecretStr | str,
    convert_secret: Callable[[str], str] | None,
) -> str: ...


@overload
def serialize_secretstr(
    data: SecretStr | str,
    convert_secret: Callable[[str], str | None],
) -> str | None: ...


@overload
def serialize_secretstr(
    data: None,
    convert_secret: Callable[[str], str | None] | Callable[[str], str] | None,
) -> None: ...


@overload
def serialize_secretstr(
    data: SecretStr | str | None,
    convert_secret: Callable[[str], str | None] | Callable[[str], str] | None = None,
) -> str | None: ...


def serialize_secretstr(
    data: SecretStr | str | None,
    convert_secret: Callable[[str], str | None] | Callable[[str], str] | None = None,
) -> str | None:
    if data is None:
        return None

    if isinstance(data, SecretStr):
        data = data.get_secret_value()

    return convert_secret(data) if convert_secret is not None else data



from ._audit import AuditPaginationFilter
from ._base import PaginationFilter
from ._order_enums import OrderBy, OrderDirection

__all__ = ["AuditPaginationFilter", "OrderBy", "OrderDirection", "PaginationFilter"]



from sqlalchemy import BigInteger
from sqlalchemy.orm import Mapped, mapped_column

from mixemy.models._base import BaseModel
from mixemy.types import ID


class IdModel(BaseModel):
    __abstract__ = True

    id: Mapped[ID] = mapped_column(BigInteger, primary_key=True, autoincrement=True)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"



from datetime import datetime

from sqlalchemy import DateTime
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.sql import func

from mixemy.models._base import BaseModel


class AuditModel(BaseModel):
    __abstract__ = True

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), server_default=func.now(), nullable=False
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), onupdate=func.now(), nullable=True
    )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} created_at={self.created_at} updated_at={self.updated_at}>"



from abc import ABC

from mixemy.repositories.sync import IdAuditSyncRepository
from mixemy.services.sync._audit import AuditSyncService
from mixemy.services.sync._id import IdSyncService
from mixemy.types import (
    CreateSchemaType,
    FilterSchemaType,
    IdAuditModelType,
    OutputSchemaType,
    UpdateSchemaType,
)


class IdAuditSyncService(
    IdSyncService[
        IdAuditModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    AuditSyncService[
        IdAuditModelType,
        CreateSchemaType,
        UpdateSchemaType,
        FilterSchemaType,
        OutputSchemaType,
    ],
    ABC,
):
    repository_type: type[IdAuditSyncRepository[IdAuditModelType]]  # pyright: ignore[reportIncompatibleVariableOverride] - https://github.com/python/typing/issues/548
    repository: IdAuditSyncRepository[IdAuditModelType]  # pyright: ignore[reportIncompatibleVariableOverride] - https://github.com/python/typing/issues/548
